//
// main.c
//
// Build Your Own Electronic Game Kit PCB
//
// This file contains code for the Microchip PIC16LF628A/648A for CCS PCM
//
// Created by Joe Grand, Grand Idea Studio, Inc., February 20, 2006
// http://www.grandideastudio.com/portfolio/build-your-own-electronic-game-kit/
//
// Distributed under a Creative Commons Attribution-Share Alike 3.0 Unported license:
// http://creativecommons.org/licenses/by-sa/3.0/
//

#include <stdlib.h>
#include <string.h>
#include "make.h"
// Per the CCS Compiler (http://www.ccsinfo.com) copyright, the source code to "tones.c" may only 
// be distributed to other licensed users of the CCS C compiler. If I had noticed that during 
// development, I would have created my own function to be included in this open source release. 
// Sorry about that! 
#include "tones.c" // we use the generate_tone function here



short sleep_mode;       // global flag to send processor into sleep mode
char  tmr0_overflow;    // counts how many times TMR0 interrupts
char  timeout_count;    // timeout delay limit
char  powerup_buttons;  // button presses on power-up
char  game_config;      // game configuration flags
char  move_number;      // move number (also address of internal EEPROM move location)
char  current_move;     // current move (same as EEPROM value so we only have to read each move from EEPROM once)


#define GAMEOVER_SND_SIZE 10
const struct note
{
   long tone;
   long length;
} gameover_snd[GAMEOVER_SND_SIZE] = {
C_note[1],400, C_note[1],50, C_note[1],100, C_note[1],500, E_note[1],100,
D_note[1],200, C_note[1],100, C_note[1],200, B_note[0],100, C_note[1],400};

#define STARTGAME_SND_SIZE 7
const struct note startgame_snd[STARTGAME_SND_SIZE] = {
F_note[1],200, D_note[1],200, F_note[1],100, D_note[1],100, G_note[1],100,
F_note[1],200, C_note[1],200};

#define WINGAME_SND_SIZE 7
const struct note wingame_snd[WINGAME_SND_SIZE] = {
C_note[2],100, Db_note[2],100, D_note[2],100, F_note[2],100, 1,100, D_note[2],50,
F_note[2],400};


const struct note red_snd[1] = {A_note[1], BETWEEN_MOVE_ON_DELAY};
const struct note green_snd[1] = {C_note[1], BETWEEN_MOVE_ON_DELAY};
const struct note blue_snd[1] = {E_note[1], BETWEEN_MOVE_ON_DELAY};
const struct note yellow_snd[1] = {G_note[1], BETWEEN_MOVE_ON_DELAY};


#INT_RTCC   // This function is called every time the RTCC (timer0) overflows
void clock_isr()
{
    if(--tmr0_overflow==0)
    {
      ++timeout_count;
      tmr0_overflow=INTS_PER_SECOND;
    }

}

#INT_RB
void rb_isr()
{
	delay_ms(25); // poor man's debounce (we shouldn't really do this in an ISR)

	if ((~(input_b() & SW_NO_BUTTON)) >> 4) // if button is still pressed
    	sleep_mode = FALSE;                 // time to wake up the processor and get to work
}

void main (void)
{
	hardware_init();
    powerup_buttons = (~(input_b() & SW_NO_BUTTON)) >> 4; // check for button presses on power-up
    enable_interrupts(GLOBAL);  // turn on interrupts

    simon_config();
    while(1)
		simon_game();
}


void simon_config (void)
{
    char i;

    // set game configuration as specified from buttons on power-up
    // red: silent mode (no tone)
    // green: blind mode (no LEDs)
    // blue: expert mode (faster speed)
    // yellow: reverse mode
    game_config = 0;
    if (powerup_buttons & SW_RED_MASK)
        game_config |= SILENT_MODE;
    if (powerup_buttons & SW_GREEN_MASK)
        game_config |= BLIND_MODE;
    if (powerup_buttons & SW_BLUE_MASK)
        game_config |= EXPERT_MODE;
    if (powerup_buttons & SW_YELLOW_MASK)
        game_config |= REVERSE_MODE;

    if (!(game_config & SILENT_MODE)) // if silent mode not selected
    {
        for (i = 0; i < STARTGAME_SND_SIZE; ++i)
        {
            generate_tone(startgame_snd[i].tone,startgame_snd[i].length);
            delay_ms(50);
        }
    }
	else
		delay_ms(1000); // delay to give the user a chance to remove their finger from the button

    while (((~(input_b() & SW_NO_BUTTON)) >> 4) != 0){}; // wait until no buttons are being pressed
}


BOOLEAN simon_game (void)
{
    char i;

    // wait for button press to start the game...
    disable_interrupts(INT_RB);  // turn off port B interrupt to avoid extra timing delays during gameplay

    // sleep() function was giving problems with power-on, so using simple polling instead
    while (((~(input_b() & SW_NO_BUTTON)) >> 4) == 0){};

    // ...then seed random number generator with current value of timer1
    srand(get_timer1());

    // reset game-specific variables
    move_number = 0;
    timeout_count = 0;
    tmr0_overflow = INTS_PER_SECOND;

    // Turn off LEDs
    output_high(LED_BLUE);
    output_high(LED_YELLOW);
    output_high(LED_RED);
    output_high(LED_GREEN);

    while (1)
    {
        disable_interrupts(INT_RTCC);
        delay_ms(1000); // allow the player to prepare before each level starts

        simon_play_moves();
        enable_interrupts(INT_RTCC);
        if (simon_read_moves()) // if bad move or timeout, end game
        {
            simon_failed_input();
            return(1);
        }

        // Move is successful
        ++move_number;
        if (move_number == EEPROM_SIZE) // maximum number of moves in a single game (set to the maximum size of internal EEPROM)
        {
            if (!(game_config & SILENT_MODE)) // if silent mode not selected
            {
                // Player has won the game!
                for (i = 0; i < WINGAME_SND_SIZE; ++i)
                {
                    generate_tone(wingame_snd[i].tone,wingame_snd[i].length);
                    delay_ms(50);
                }
            }
            else    // if it is, just turn on all the LEDs
            {
                output_low(LED_BLUE);
                output_low(LED_YELLOW);
                output_low(LED_RED);
                output_low(LED_GREEN);
            }

            return(1);   // loop back to main routine and wait for a keypress to start a new game
        }
    }
}


void simon_play_moves(void)
{
    char i;

    current_move = rand() & 0x03;
    write_eeprom(move_number, current_move); // store current move in internal EEPROM and add to the end of the pattern

    for (i = 0; i <= move_number; ++i) // play all moves for the current level
    {
        current_move = read_eeprom(i);

        switch (current_move)
        {
            case RED_MOVE:
                if (!(game_config & BLIND_MODE))
                    output_low(LED_RED);
                if (!(game_config & SILENT_MODE))
                    generate_tone(red_snd[0].tone,red_snd[0].length);
                break;
            case GREEN_MOVE:
                if (!(game_config & BLIND_MODE))
                    output_low(LED_GREEN);
                if (!(game_config & SILENT_MODE))
                    generate_tone(green_snd[0].tone,green_snd[0].length);
                break;
            case BLUE_MOVE:
                if (!(game_config & BLIND_MODE))
                    output_low(LED_BLUE);
                if (!(game_config & SILENT_MODE))
                    generate_tone(blue_snd[0].tone,blue_snd[0].length);
                break;
            case YELLOW_MOVE:
                if (!(game_config & BLIND_MODE))
                    output_low(LED_YELLOW);
                if (!(game_config & SILENT_MODE))
                    generate_tone(yellow_snd[0].tone,yellow_snd[0].length);
                break;
        }

        // if silent mode enabled, need to force an on-time delay (instead of relying on generate_tone)
        if (game_config & SILENT_MODE)
            delay_ms(BETWEEN_MOVE_ON_DELAY);

        // Turn off LEDs
        output_high(LED_BLUE);
        output_high(LED_YELLOW);
        output_high(LED_RED);
        output_high(LED_GREEN);

        // delay to wait before showing the next move
        if (game_config & EXPERT_MODE)  // if expert mode was selected...
            delay_ms(BETWEEN_MOVE_OFF_DELAY_HARD); // ...then set the speed accordingly
        else
            delay_ms(BETWEEN_MOVE_OFF_DELAY_EASY);
    }
}


BOOLEAN simon_read_moves(void)
{
    char i, but, timeout;

    if (game_config & EXPERT_MODE)  // if expert mode was selected...
        timeout = TIMEOUT_DELAY_HARD; // ...then set the speed accordingly
    else
        timeout = TIMEOUT_DELAY_EASY;

    for (i = 0; i <= move_number; ++i)  // read each move stored in the EEPROM and compare it against the player's input
    {
        if (game_config & REVERSE_MODE) // if reverse mode was selected...
            current_move = read_eeprom(move_number - i);  // reverse the pointers so the last move needs to be entered by the player first (moving backwards)
        else
            current_move = read_eeprom(i);

        timeout_count = 0;

        while (timeout_count != timeout)  // wait for player to repeat moves
        {
            // poll for the correct keypress
            but = (~(input_b() & SW_NO_BUTTON)) >> 4; // check for port b button presses
            if (but)
            {
                delay_ms(25); // poor man's debounce
                but = (~(input_b() & SW_NO_BUTTON)) >> 4;

                // if correct, play sound and show LED of current button press
                if ((but & SW_RED_MASK) && (current_move == RED_MOVE))
                {
                    timeout_count = 0;
                    if (!(game_config & BLIND_MODE))
                        output_low(LED_RED);
                    if (!(game_config & SILENT_MODE))
                        generate_tone(red_snd[0].tone,red_snd[0].length);
                    else
                        delay_ms(BETWEEN_MOVE_ON_DELAY);
                    while(!input(SW_RED))    // wait until player releases button
                    {
                        if (timeout_count == timeout) return(1); // prevent player from holding down button for too long
                    }
                    goto next;
                }
                else if ((but & SW_GREEN_MASK) && (current_move == GREEN_MOVE))
                {
                    timeout_count = 0;
                    if (!(game_config & BLIND_MODE))
                        output_low(LED_GREEN);
                    if (!(game_config & SILENT_MODE))
                        generate_tone(green_snd[0].tone,green_snd[0].length);
                    else
                        delay_ms(BETWEEN_MOVE_ON_DELAY);
                    while(!input(SW_GREEN))    // wait until player releases button
                    {
                        if (timeout_count == timeout) return(1); // prevent player from holding down button for too long
                    }
                    goto next;
                }
                else if ((but & SW_BLUE_MASK) && (current_move == BLUE_MOVE))
                {
                    timeout_count = 0;
                    if (!(game_config & BLIND_MODE))
                        output_low(LED_BLUE);
                    if (!(game_config & SILENT_MODE))
                        generate_tone(blue_snd[0].tone,blue_snd[0].length);
                    else
                        delay_ms(BETWEEN_MOVE_ON_DELAY);
                    while(!input(SW_BLUE))    // wait until player releases button
                    {
                        if (timeout_count == timeout) return(1); // prevent player from holding down button for too long
                    }
                    goto next;
                }
                else if ((but & SW_YELLOW_MASK) && (current_move == YELLOW_MOVE))
                {
                    timeout_count = 0;
                    if (!(game_config & BLIND_MODE))
                        output_low(LED_YELLOW);
                    if (!(game_config & SILENT_MODE))
                        generate_tone(yellow_snd[0].tone,yellow_snd[0].length);
                    else
                        delay_ms(BETWEEN_MOVE_ON_DELAY);
                    while(!input(SW_YELLOW))    // wait until player releases button
                    {
                        if (timeout_count == timeout) return(1); // prevent player from holding down button for too long
                    }
                    goto next;
                }
                else
                    return(1);
            }
        }
        return(1); // if we get here, the user has timed out

        // code will jump to here if a correct button press was made, so continue onto the next move
next:
        // Turn off LEDs
        output_high(LED_BLUE);
        output_high(LED_YELLOW);
        output_high(LED_RED);
        output_high(LED_GREEN);
    }

    return(0);
}


// show the LED that was supposed to be entered and sound the "game over" music
void simon_failed_input(void)
{
    char i;
    char hundreds, tens, ones;

    switch (current_move)
    {
        case RED_MOVE:
            output_high(LED_BLUE);
            output_high(LED_YELLOW);
            output_low(LED_RED);
            output_high(LED_GREEN);
            break;
        case GREEN_MOVE:
            output_high(LED_BLUE);
            output_high(LED_YELLOW);
            output_high(LED_RED);
            output_low(LED_GREEN);
            break;
        case BLUE_MOVE:
            output_low(LED_BLUE);
            output_high(LED_YELLOW);
            output_high(LED_RED);
            output_high(LED_GREEN);
            break;
        case YELLOW_MOVE:
            output_high(LED_BLUE);
            output_low(LED_YELLOW);
            output_high(LED_RED);
            output_high(LED_GREEN);
            break;
    }

    if (!(game_config & SILENT_MODE)) // if silent mode not selected
    {
        for (i = 0; i < GAMEOVER_SND_SIZE; ++i)
        {
            generate_tone(gameover_snd[i].tone,gameover_snd[i].length);
            delay_ms(50);
        }
    }

    delay_ms(1500);

    // turn off LEDs
    output_high(LED_BLUE);
    output_high(LED_YELLOW);
    output_high(LED_RED);
    output_high(LED_GREEN);

    delay_ms(1500); // pause before displaying the score

    hundreds = move_number / 100;
    tens = (move_number % 100) / 10;
    ones = (move_number % 10);

    // blink the score using green (hundreds), red (tens), and blue (ones) LEDs
    for (i = 0; i < hundreds; ++i)
    {
        output_low(LED_GREEN);
        if (!(game_config & SILENT_MODE))
            generate_tone(green_snd[0].tone,green_snd[0].length);
        else
            delay_ms(BETWEEN_MOVE_ON_DELAY);
        output_high(LED_GREEN);
        delay_ms(BETWEEN_MOVE_OFF_DELAY_EASY);
    }

    for (i = 0; i < tens; ++i)
    {
        output_low(LED_RED);
        if (!(game_config & SILENT_MODE))
            generate_tone(red_snd[0].tone,red_snd[0].length);
        else
            delay_ms(BETWEEN_MOVE_ON_DELAY);
        output_high(LED_RED);
        delay_ms(BETWEEN_MOVE_OFF_DELAY_EASY);
    }

    for (i = 0; i < ones; ++i)
    {
        output_low(LED_BLUE);
        if (!(game_config & SILENT_MODE))
            generate_tone(blue_snd[0].tone,blue_snd[0].length);
        else
            delay_ms(BETWEEN_MOVE_ON_DELAY);
        output_high(LED_BLUE);
        delay_ms(BETWEEN_MOVE_OFF_DELAY_EASY);
    }
}


void hardware_init (void) // Configure hardware to a known state
{
    // Enable port B pull-ups for switch inputs
    port_b_pullups(PIN_B4);
    port_b_pullups(PIN_B5);
    port_b_pullups(PIN_B6);
    port_b_pullups(PIN_B7);

    // Turn off LEDs
    output_high(LED_BLUE);
    output_high(LED_YELLOW);
    output_high(LED_RED);
    output_high(LED_GREEN);

    // Setup timer1 for LED flashing
    setup_timer_1(T1_INTERNAL | T1_DIV_BY_8);
    set_timer1(0);

    // Setup timer0 for timeout counting
    set_timer0(0);
    setup_counters(RTCC_INTERNAL, RTCC_DIV_256 | RTCC_8_BIT);
    enable_interrupts(INT_RTCC);

    sleep_mode = TRUE;          // init sleep flag
  	ext_int_edge(H_TO_L);       // init interrupt triggering for button press
  	enable_interrupts(INT_RB);  // enable interrupt on Port B4-B7 any change
}
